

#include <tcl.h>
#include <iostream>
#include <string>
#include <string.h>
#include <stdlib.h> 
#include <itcl.h>
#include <vector>
#include <sstream>

using std::cout;
using std::endl;

class Collection {
    public:
        std::string collection_name;
        int index;
        Tcl_Command token;
        Collection() : token(NULL) {}
};

// Implementation guts of the collection
int Do_Tcl_Collection_Command( ClientData clientData, Tcl_Interp* interp, int argc, Tcl_Obj* CONST argv[] ) {
    Collection* f = static_cast<Collection*>(clientData);
    if ( argc == 2 ) {
        std::string cmd = Tcl_GetStringFromObj( argv[1], NULL );
        if ( cmd == "name" ) {
            Tcl_Obj* result = Tcl_NewStringObj( f->collection_name.c_str(), -1 );
            Tcl_SetObjResult( interp, result );
            return TCL_OK;
        }
    } else {
        return TCL_ERROR;
    }
}

// Delete a collection command object
void Delete_Tcl_Collection( ClientData clientData ) {
    Collection* f = static_cast<Collection*>(clientData);
    cout << "Called Delete_Tcl_Collection on index " << f->index << endl;
    delete f;
}

// Create a collection command object
Collection* Create_Tcl_Collection ( Tcl_Interp* interp, std::string& my_command_name, std::string& my_given_name, int collection_index )
{
    Collection* f = new Collection;
    f->index = collection_index;
    f->collection_name = my_given_name;
    Tcl_Command cmdPtr = Tcl_CreateObjCommand( 
            interp, 
            my_command_name.c_str(), 
            Do_Tcl_Collection_Command, 
            static_cast<void*>(f), 
            Delete_Tcl_Collection );
    f->token = cmdPtr;
    return f;
}

// Keep track of all collections
static std::vector<Collection*> AllCollections;

// A Tcl command that allows you to fiddle with the collection objects
int collection_manager(
    ClientData c_data,
    Tcl_Interp *interp,
    int argc,
    Tcl_Obj *     CONST argv[]
)
{
    const char* usage = "usage: collection_manager collection_obj collection_command";
    if ( argc != 3 ) {
        cout << usage << endl;
        return TCL_ERROR;
    }
    std::string arg1 = Tcl_GetStringFromObj( argv[1], NULL );
    if ( arg1 == "make" ) {
        int i=0;
        std::string arg2 = Tcl_GetStringFromObj( argv[2], NULL );
        size_t max_index = AllCollections.size();
        for( ; i<max_index && AllCollections[i]!=NULL; ++i );
        if ( i == max_index ) {
            // Need a new slot
            AllCollections.push_back(NULL);
        }

        std::stringstream collection_name;
        collection_name << "mystuff::Collection" << i;
        std::string the_collection = collection_name.str();
        std::string the_collection_name = arg2;
        AllCollections[i] = Create_Tcl_Collection(interp, the_collection, the_collection_name, i);
        Tcl_Obj* result_string = Tcl_NewStringObj( the_collection.c_str(), the_collection.length() );
        Tcl_SetObjResult( interp, result_string );
        return TCL_OK;
    }
    std::string command_name = Tcl_GetStringFromObj( argv[2], NULL );
    // Identify the Collection command object
    Tcl_CmdInfo Collection_obj;
    int rtval = Tcl_GetCommandInfo( interp, arg1.c_str(), &Collection_obj );
    if ( ! rtval ) {
        cout << "Not a collection." << endl;
        return TCL_ERROR;
    }
    // Ideally, one might check things out a bit here first...
    //
    if ( command_name == "name" ) {
        // Pass through to the object to ask its name by calling the underlying command directly
        return Collection_obj.objProc( Collection_obj.objClientData, interp, argc-1, &argv[1] );
    } else if ( command_name == "delete" ) {
        Collection* the_collection = static_cast<Collection*>(Collection_obj.objClientData); 
        Tcl_Command collection_token = the_collection->token;
        AllCollections[the_collection->index] = NULL;
        int rtval = Tcl_DeleteCommandFromToken( interp, collection_token );

        if ( rtval == -1 ) {
            cout << "Not a collection." << endl;
            return TCL_ERROR;
        } else {
            return TCL_OK;
        }
    } else {
        cout << "bad command" << endl;
        return TCL_ERROR;
    }

    return TCL_OK;
}

int Collection_Appinit( Tcl_Interp* interp ) {
    if ((Tcl_Init)(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    Tcl_CreateObjCommand(interp, "collection_manager", collection_manager, NULL, NULL);
    return TCL_OK;
}

int main( int argc, char* argv[] ) {

    Tcl_Main( argc, argv, &Collection_Appinit );

}
